@@ -26,23 +26,25 @@ Features
 * Ability to authenticate using API key
 * In-built and cusomizable limit to protect against frequent API calls
 * Local file-based storage for frequent requests (persistent cache)
-
-
-@author: Hrishikesh Terdalkar
 """
 
 import os
 import time
 import json
 import logging
+from typing import Any, List
 from urllib.parse import urlsplit, urlunsplit
+from dataclasses import dataclass, field
 
-import attr
 import requests
 
 ###############################################################################
 
-logger = logging.getLogger(__name__)
+LOGGER = logging.getLogger(__name__)
+
+###############################################################################
+
+OPENDOTA_API_URL = "https://api.opendota.com/api"
 
 ###############################################################################
 
@@ -56,7 +58,7 @@ FANTASY = {
     'assists': 0.15,
     'last_hits': 0.003,
     'gold_per_min': 0.002,
-    'xp_per_min': 0.002,
+    'xp_per_min': 0,
     'tower_kills': 1,
     'roshan_kills': 1,
     'teamfight': 3,
@@ -64,8 +66,8 @@ FANTASY = {
     'camps_stacked': 0.5,
     'runes_grabbed': 0.25,
     'first_blood': 4,
-    'stuns': 0.07,
-    'hero_healing': 0.0004,
+    'stuns': 0.05,
+    'hero_healing': 0,
 
     # Add '_base' suffix to the parameter for specifying base score
     'deaths_base': 3,
@@ -74,22 +76,22 @@ FANTASY = {
 ###############################################################################
 
 
-@attr.s
+@dataclass
 class OpenDota:
     """
     <OPENDOTA/> API Interface
 
     Instance of a connection to OpenDota API.
     All methods return serializable python objects, which are also stored
-    as JSON in the `data_dir` for future calls.
-    All methods take a boolean argument `force` which, if `True`, will fetch
-    the data again even if it is available in the data directory.
+    as JSON in the :code:`data_dir` for future calls.
+    All methods take a boolean argument :code:`force` which, if True,
+    will fetch the data again even if it is available in the data directory.
 
     Parameters
     ----------
         data_dir: str, (optional)
             Path to data directory for storing responses to API calls
-            The default is ~/dota2.
+            The default is :code:`~/dota2`.
         api_key: str, (optional)
             If you have an OpenDota API key
             The default is None.
@@ -103,38 +105,51 @@ class OpenDota:
             Fantasy DotA2 Configuration
             Utility constant FANTASY holds the standard values
             and is used as default.
-            Keys of the fantasy must match those of FANTASY.
+            Keys of the :code:`fantasy` will override the default values.
+            They must be a subset of the keys of :code:`FANTASY`.
 
-            Parameters ending with `_base` are used as base values,
+            Parameters ending with :code:`_base` are used as base values,
             while others are used as multipliers.
             e.g.
-                `deaths` = -0.3
-                `deaths_base` = 3
-            results in the calculation as follows,
-                `death_score` = 3 + (number_of_deaths * -0.3)
-            If `_base` parameter is absent, it's assumed to be 0.
+            :code:`deaths = -0.3` and :code:`deaths_base = 3` results in the
+            calculation as, :code:`death_score = 3 + (number_of_deaths * -0.3)`
+            If :code:`_base` parameter is absent, it's assumed to be 0.
         api_url: str, (optional)
             URL to OpenDota API.
             It is recommended to not change this value.
     """
 
-    data_dir: str = attr.ib(default=None)
-    api_key: str = attr.ib(default=None, repr=False)
-    delay: int = attr.ib(default=3, repr=False)
-    fantasy: dict = attr.ib(default=FANTASY, repr=False)
-    api_url: str = attr.ib(default="https://api.opendota.com/api", repr=False)
+    data_dir: str = field(default=None)
+    api_key: str = field(default=None, repr=False)
+    delay: int = field(default=3, repr=False)
+    fantasy: dict = field(default=None, repr=False)
+    api_url: str = field(default=OPENDOTA_API_URL, repr=False)
 
-    def __attrs_post_init__(self):
+    def __post_init__(self):
         self._session = requests.Session()
         if self.api_key is not None:
             self._session.headers['Authorization'] = f'Bearer {self.api_key}'
+
         if self.data_dir is None:
             self.data_dir = os.path.join(os.path.expanduser("~"), "dota2")
         os.makedirs(self.data_dir, exist_ok=True)
 
+        default_fantasy = FANTASY.copy()
+        if self.fantasy is not None:
+            default_fantasy.update(self.fantasy)
+        self.fantasy = default_fantasy
+
     # ----------------------------------------------------------------------- #
 
-    def request(self, url, *, post=False, data={}, filename=None, force=False):
+    def request(
+        self,
+        url: str,
+        *,
+        post: bool = False,
+        data: dict = None,
+        filename: str = None,
+        force: bool = False
+    ) -> Any:
         """
         Make a GET or POST request to <OPENDOTA/> API
 
@@ -147,7 +162,7 @@ class OpenDota:
                 The default is False.
             data: dict, (optional)
                 Query Data.
-                The default is {}.
+                The default is None.
             filename: str, (optional)
                 Save the data to this file.
                 The default is None.
@@ -160,7 +175,6 @@ class OpenDota:
             object:
                 Result of the API call deserialized as a python object
         """
-
         path = None
         if filename is not None:
             path = os.path.join(self.data_dir, filename)
@@ -168,13 +182,14 @@ class OpenDota:
                 try:
                     with open(path, "r", encoding="utf-8") as f:
                         json_data = json.load(f)
-                    logger.info(
+                    LOGGER.info(
                         f"Loading previously fetched data from '{filename}'."
                     )
                     return json_data
                 except Exception:
                     pass
 
+        data = data or {}
         if self.api_key is None:
             time.sleep(self.delay)
         else:
@@ -186,7 +201,7 @@ class OpenDota:
             url_parts[3] = "&".join([f"{k}={v}" for k, v in data.items()])
 
         query_url = urlunsplit(url_parts)
-        logger.info("Query URL:", query_url)
+        LOGGER.info("Query URL:", query_url)
 
         if not post:
             r = self._session.get(query_url)
@@ -197,7 +212,7 @@ class OpenDota:
         json_data = json.loads(content)
 
         if json_data and "error" in json_data:
-            logger.warning(f"Could not fetch '{url}' ({json_data['error']}).")
+            LOGGER.warning(f"Could not fetch '{url}' ({json_data['error']}).")
             return None
 
         if path is not None:
@@ -226,13 +241,13 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Request
 
-    def request_parse(self, match_id):
+    def request_parse(self, match_id: int or str):
         """Submit a new parse request"""
-        logger.info(f"Requesting parse for match {match_id}")
+        LOGGER.info(f"Requesting parse for match {match_id}")
         url = f"/request/{match_id}"
         return self.post(url)
 
-    def request_status(self, job_id):
+    def request_status(self, job_id: int or str):
         """Get parse request state"""
         url = f"/request/{job_id}"
         return self.get(url)
@@ -240,26 +255,27 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Static Game Data (mirrored from the dotaconstants repository)
 
-    def get_constant_names(self, force=False):
+    def get_constant_names(self, force: bool = False):
         """Get an array of available resources"""
         url = "/constants"
         filename = "constants.json"
         return self.get(url, filename=filename, force=force)
 
-    def get_constants(self, resource=None, force=False):
+    def get_constants(self, resource: str = None, force: bool = False):
         """
         Get static game data for specified resource(s)
         (mirrored from the dotaconstants repository)
 
         Parameters
         ----------
-            resource: str or list, indicating name or names of resources
+            resource: str or list,
+                Name or names of resources
         """
         if resource is not None:
             if isinstance(resource, str):
                 resource = [resource]
             if not isinstance(resource, list):
-                logger.error(
+                LOGGER.error(
                     "`resources' must be a string or a list of strings, "
                     f"not `{type(resource)}'"
                 )
@@ -277,19 +293,19 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Hero
 
-    def get_heroes(self, force=False):
+    def get_heroes(self, force: bool = False):
         """Get hero data"""
         url = "/heroes"
         filename = "heroes.json"
         return self.get(url, filename=filename, force=force)
 
-    def get_hero_stats(self, force=False):
+    def get_hero_stats(self, force: bool = False):
         """Get stats about hero performance in recent matches"""
         url = "/heroStats"
         filename = "hero_stats.json"
         return self.get(url, filename=filename, force=force)
 
-    def get_hero_benchmarks(self, hero_id, force=False):
+    def get_hero_benchmarks(self, hero_id: int or str, force: bool = False):
         """Get benchmarks for a hero"""
         url = "/benchmarks"
         filename = f"benchmarks_{hero_id}.json"
@@ -299,13 +315,13 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Leagues
 
-    def get_leagues(self, force=False):
+    def get_leagues(self, force: bool = False):
         """Get a list of leagues"""
         url = "/leagues"
         filename = "leagues.json"
         return self.get(url, filename=filename, force=force)
 
-    def get_league(self, league_id, force=False):
+    def get_league(self, league_id: int or str, force: bool = False):
         """Get data for a league"""
         url = f"/leagues/{league_id}"
         filename = f"league_{league_id}.json"
@@ -314,13 +330,13 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # League Specific
 
-    def get_league_matches(self, league_id, force=False):
+    def get_league_matches(self, league_id: int or str, force: bool = False):
         """Get matches from a league"""
         url = f"/leagues/{league_id}/matches"
         filename = f"league_{league_id}_matches.json"
         return self.get(url, filename=filename, force=force)
 
-    def get_league_teams(self, league_id, force=False):
+    def get_league_teams(self, league_id: int or str, force: bool = False):
         """Get teams from a league"""
         url = f"/leagues/{league_id}/teams"
         filename = f"league_{league_id}_teams.json"
@@ -329,7 +345,7 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Teams
 
-    def get_teams(self, force=False):
+    def get_teams(self, force: bool = False):
         """Get team data"""
         url = "/teams"
         filename = "teams.json"
@@ -342,7 +358,7 @@ class OpenDota:
                 json.dump(team, f, ensure_ascii=False)
         return teams
 
-    def get_team(self, team_id, force=False):
+    def get_team(self, team_id: int or str, force: bool = False):
         """Get data for a team"""
         url = f"/teams/{team_id}"
         filename = f"team_{team_id}.json"
@@ -351,13 +367,18 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Team Specific
 
-    def get_team_matches(self, team_id, force=False):
+    def get_team_matches(self, team_id: int or str, force: bool = False):
         """Get matches for a team"""
         url = f"/teams/{team_id}/matches"
         filename = f"team_{team_id}_matches.json"
         return self.get(url, filename=filename, force=force)
 
-    def get_team_players(self, team_id, current=True, force=False):
+    def get_team_players(
+        self,
+        team_id: int or str,
+        current: bool = True,
+        force: bool = False
+    ):
         """Get players who have played for a team"""
         url = f"/teams/{team_id}/players"
         filename = f"team_{team_id}_players.json"
@@ -371,7 +392,7 @@ class OpenDota:
         else:
             return players
 
-    def get_team_heroes(self, team_id, force=False):
+    def get_team_heroes(self, team_id: int or str, force: bool = False):
         """Get heroes for a team"""
         url = f"/teams/{team_id}/heroes"
         filename = f"team_{team_id}_heroes.json"
@@ -380,13 +401,13 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Matches
 
-    def get_match(self, match_id, force=False):
+    def get_match(self, match_id: int or str, force: bool = False):
         """Get match data"""
         url = f"/matches/{match_id}"
         filename = f"match_{match_id}.json"
         return self.get(url, filename=filename, force=force)
 
-    def get_pro_matches(self, force=False):
+    def get_pro_matches(self, force: bool = False):
         """Get a list of pro matches"""
         url = "/proMatches"
         filename = "pro_matches.json"
@@ -400,13 +421,13 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Players
 
-    def get_player(self, account_id, force=False):
+    def get_player(self, account_id: int or str, force: bool = False):
         """Player data"""
         url = f"/players/{account_id}"
         filename = f"player_{account_id}.json"
         return self.get(url, filename=filename, force=force)
 
-    def get_pro_players(self, force=False):
+    def get_pro_players(self, force: bool = False):
         """Get a list of pro players"""
         url = "/proPlayers"
         filename = "pro_players.json"
@@ -415,7 +436,7 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Player Specific
 
-    def get_player_heroes(self, player_id, force=False):
+    def get_player_heroes(self, player_id: int or str, force: bool = False):
         """Get heroes played by a player"""
         url = f"/players/{player_id}/heroes"
         filename = f"player_{player_id}_heroes.json"
@@ -429,7 +450,11 @@ class OpenDota:
         return self.get(url, filename=filename, data=data, force=force)
 
     def get_player_matches(
-        self, player_id, request_parse=False, days=180, force=False
+        self,
+        player_id: int or str,
+        request_parse: bool = False,
+        days: int = 180,
+        force: bool = False
     ):
         """Matches played by a player"""
         url = f"/players/{player_id}/matches"
@@ -441,17 +466,17 @@ class OpenDota:
                 match_id = match["match_id"]
                 if match["version"] is None or match["version"] < 20:
                     json_data = self.request_parse(match_id)
-                    logger.info("Match ID:", match_id)
-                    logger.info("Job ID:", json_data["job"]["jobId"])
+                    LOGGER.info("Match ID:", match_id)
+                    LOGGER.info("Job ID:", json_data["job"]["jobId"])
         return matches
 
-    def get_player_ratings(self, player_id, force=False):
+    def get_player_ratings(self, player_id: int or str, force: bool = False):
         """Player rating history"""
         url = f"/players/{player_id}/ratings"
         filename = f"player_{player_id}_ratings.json"
         return self.get(url, filename=filename, force=force)
 
-    def get_player_rankings(self, player_id, force=False):
+    def get_player_rankings(self, player_id: int or str, force: bool = False):
         """Player hero rankings"""
         url = f"/players/{player_id}/rankings"
         filename = f"player_{player_id}_rankings.json"
@@ -460,7 +485,12 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Search
 
-    def search_hero(self, search_key=None, attack_type=None, roles=None):
+    def search_hero(
+        self,
+        search_key: str = None,
+        attack_type: str = None,
+        roles: List[str] = None
+    ):
         """Search for a hero by name, attack type or roles"""
         results = []
         for hero in self.get_heroes():
@@ -481,7 +511,7 @@ class OpenDota:
                 results.append(hero)
         return results
 
-    def search_league(self, search_key):
+    def search_league(self, search_key: str):
         """Search for a league"""
         return [
             league
@@ -489,7 +519,7 @@ class OpenDota:
             if search_key.lower() in league["name"].lower()
         ]
 
-    def search_team(self, search_key):
+    def search_team(self, search_key: str):
         """Search for a team by name or tag"""
         return [
             team
@@ -498,7 +528,7 @@ class OpenDota:
             or search_key.lower() == team["tag"].lower()
         ]
 
-    def search_player(self, search_key):
+    def search_player(self, search_key: str):
         """Search for a player"""
         url = "/search"
         data = {"q": search_key}
@@ -507,10 +537,16 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Fantasy
 
-    def get_fantasy_points(self, match_id, force=False):
+    def get_fantasy_points(self, match_id: int or str, force: bool = False):
         """
         Get Fantasy Points of All Players from a Match
 
+        Parameters
+        ----------
+
+        match_id: int or str
+            Match ID
+
         Returns
         -------
             list:
@@ -588,7 +624,7 @@ class OpenDota:
     # ----------------------------------------------------------------------- #
     # Database
 
-    def get_schema(self, table_name=None, force=False):
+    def get_schema(self, table_name: str = None, force: bool = False):
         """
         Get database schema
 
@@ -611,7 +647,7 @@ class OpenDota:
                 if table_name == field["table_name"]
             }
 
-    def explorer(self, sql, debug=False):
+    def explorer(self, sql: str, debug: bool = False):
         """Submit arbitrary PostgreSQL queries to the database"""
         url = "/explorer"
         data = {"sql": sql}
@@ -626,7 +662,7 @@ class OpenDota:
 
     # ----------------------------------------------------------------------- #
 
-    def update_data(self, frequency=FREQ_HIGH):
+    def update_data(self, frequency: int = FREQ_HIGH):
         """
         Update core data
 
